import re
import os
import numpy as np
import tkinter
import tomllib
import multiprocessing as mp
from tkinter.filedialog import askopenfilename, askopenfilenames

# Default .csv file for polynomial coefficents of fit for motor steps vs angle (see SPR_poly_coefficients_generator.py).
# This should be regenerated when used in a new instrument. Does generally not need to be regenerated after normal calibrations,
# unless the stepper motor calibration in the instrument has changed for some reason (e.g. may happen from homing errors or similar).
# Assumes all selected measurement files comes from the same instrument. Use multiple copies of spr2_to_csv_v2.py and
# their associated SPR_poly_coeff_YY-MM-DD.csv file if working with multiple instruments.


# Read configuration parameters
with open('config.toml', 'r') as f:
    config = tomllib.loads(f.read())

default_poly_file = config["default_poly_file"]
max_logical_cores = config["max_logical_cores"]

# Determine how many processes can be used for calculations at a time
if max_logical_cores == 0:
    logical_cores = mp.cpu_count()
elif max_logical_cores > mp.cpu_count():
    print('Warning: max_logical_cores exceeding system specifications. Using all available cores.')
    logical_cores = mp.cpu_count()
else:
    logical_cores = max_logical_cores


def extract_parameters(content):

    # Calibration data
    calibration_device_pattern = re.compile(r'<calibration device_serial=.*>')
    cal_channels_pattern = re.compile(r'channels="\d+"')
    cal_step_pattern = re.compile(r'step_len="\d+"')
    cal_points_pattern = re.compile(r'points="\d+"')
    cal_start_pos_pattern = re.compile(r'start_pos="\d+"')
    digit_pattern = re.compile(r'\d+')

    calibration_param_line = calibration_device_pattern.findall(content)[0]
    channels = int(digit_pattern.findall(cal_channels_pattern.findall(calibration_param_line)[0])[0])
    cal_scanspeed = int(digit_pattern.findall(cal_step_pattern.findall(calibration_param_line)[0])[0])
    cal_points = int(digit_pattern.findall(cal_points_pattern.findall(calibration_param_line)[0])[0])
    cal_start_pos = int(digit_pattern.findall(cal_start_pos_pattern.findall(calibration_param_line)[0])[0])

    #  TIR steps for each laser in calibration
    TIR_pattern = re.compile(r'<prm_set set="P1500">.*</prm_set>')
    TIR_line = TIR_pattern.findall(content)[0]

    # set P1500 - Laser channels start at index pos 15 (starting from 0), ends at 22
    # TIR angle calibration step values starts at 26 and ends at 33
    laser_channels = list(map(int, TIR_line.split(';')[15+(8-channels):23]))

    #  Starting position for measurement scans
    pos_pattern1 = re.compile(r'<ch number="\d" start_pos="\d+">')
    pos_line = pos_pattern1.findall(content)
    pos_pattern2 = re.compile(r'"\d+">')
    pos_value = pos_pattern2.search(pos_line[-1]).group()
    start_pos = int(pos_value.strip('">'))

    #  List of measurement time for each point
    time_pattern = re.compile(r'<scan rtime="\d+"')
    time_matches = time_pattern.findall(content)
    time_value_list = [0]*len(time_matches)
    time_value_match = re.compile(r'\d+')

    for t_ind, time_string in enumerate(time_matches):
        time_value_list[t_ind] = int(time_value_match.search(time_string).group())/1000/60

    #  Determine the scanning speed/step length
    step_pattern = re.compile(r'<scan rtime="\d+" step_len="\d{1,2}" dir="Forward">')
    step_string = step_pattern.search(content).group()
    step_length_pattern = re.compile(r'"\d{1,2}"')
    step_length = int(step_length_pattern.search(step_string).group().strip('"'))

    print('Start position: ', start_pos)
    print('Scan speed: ', step_length)

    return start_pos, step_length, channels, laser_channels, time_value_list, cal_scanspeed, cal_points, cal_start_pos


def extract_spectra(content, c_ind, polycoff, start_pos, scanspeed, cal_scanspeed, cal_points, cal_start_pos, time_values, laser_channels, spr2_file):
    #  Extracts and calibrates spectra from .sp2 file, then saves it as .csv

    #  Get the spectra data (angles and intensity)
    spectra_pattern = re.compile(r'<ch number="' + str(c_ind) + r'" start_pos="\d+">.*</ch>')
    channel = spectra_pattern.findall(content)

    # Remove init_scan if it is not the only measurement
    init_scan_pattern = re.compile(r'<init_scan rtime="[1-9]')
    init_scan_match = re.search(init_scan_pattern, content)
    if init_scan_match and len(channel) > 1:
        channel.pop(0)

    data_pattern = re.compile(r'>.*<')
    point_string = data_pattern.search(channel[-1]).group().strip('><')
    points = len(point_string.split(';'))
    spectra_array = np.ones((len(channel), points))

    for row, match in enumerate(channel):
        spectra = data_pattern.search(match).group().strip('><')
        try:
            spectra_array[row, :] = list(map(float, spectra.split(';')))
        except ValueError:
            print('Row ', str(row), ': Mismatch in angular resolution compared to last scan. This scan will be skipped.')
            continue

    # Generation of angles and combining with spectra
    spectra_steps = np.arange(start_pos, (scanspeed * points) + start_pos, scanspeed)
    spectra_angles = np.polyval(polycoff, spectra_steps)
    spectra_full_array = np.vstack((spectra_angles, spectra_array))

    #  Get the calibration data
    calib_channel_pattern = re.compile(r'<number>' + str(c_ind) + r'.*</data>', flags=re.DOTALL)
    calib_channel_string = calib_channel_pattern.search(content).group()

    calib_data_pattern = re.compile(r'a>.*<')
    calib_data_string = calib_data_pattern.search(calib_channel_string).group()

    calib_data_string = calib_data_string.strip('a>')
    calib_data_string = calib_data_string.strip('<')

    calib_data = np.array(list(map(float, calib_data_string.split(';'))))/10000
    calib_steps = np.arange(float(cal_start_pos), float(cal_scanspeed*cal_points)+float(cal_start_pos), cal_scanspeed)
    calib_angles = np.polyval(polycoff, calib_steps)
    calib_array = np.vstack((calib_angles, calib_data))

    #  Start intensity calibration
    for a_ind, angle in enumerate(spectra_full_array[0, :]):
        try:
            cal_ind = next(ind1 for ind1, cal_angle in enumerate(calib_array[0, :]) if np.isclose(cal_angle, angle, 0.001))  # Search for calibration value close to its respective spectra value at a given angle
            spectra_full_array[1:, a_ind] = np.true_divide(spectra_full_array[1:, a_ind], calib_array[1, cal_ind])   # Divide all values in  spectra_full_array[1:, a_ind] with calib_array[1, cal_ind]
        except:
            try:
                # Try wider tolerance
                cal_ind = next(ind1 for ind1, cal_angle in enumerate(calib_array[0, :]) if np.isclose(cal_angle, angle, 0.003))  # Search for calibration value close to its respective spectra value at a given angle
                spectra_full_array[1:, a_ind] = np.true_divide(spectra_full_array[1:, a_ind], calib_array[1, cal_ind])  # Divide all values in  spectra_full_array[1:, a_ind] with calib_array[1, cal_ind]
            except:
                print('WARNING: Angle ', str(angle), ' was not calibrated')
                continue

    #  Add time values as first column
    time_values_np = np.array(time_values)
    spectra_full_array = np.column_stack((np.insert(time_values_np, 0, 0), spectra_full_array))

    #  Save data as .csv
    spr2_path, spr2_file_name = os.path.split(spr2_file)
    file_identifier = '{head}-L{index}_{wavelength}nm.csv'.format(head=spr2_file_name[:-5], index=str(c_ind+1), wavelength=str(laser_channels[c_ind]))
    save_name = os.path.join(spr2_path, file_identifier)
    header_string = 'Left most column is Time (min), First row is Angles (deg), Scanspeed=' + str(scanspeed)

    np.savetxt(save_name, spectra_full_array, fmt='%1.6f', delimiter=';', header=header_string)


if __name__ == '__main__':  # This is important since mp.Process goes through this file for extract_spectra()
    tkinter.Tk().withdraw()

    spr2_files = askopenfilenames(title='Select spr2 files')

    for file_ind, spr2_file in enumerate(spr2_files):

        spr2_path, spr2_file_name = os.path.split(spr2_file)

        #  Read sp2 file
        with open(spr2_file, 'r') as f:
            content = f.read()

        #  Get various parameters from file
        start_pos, scan_speed, channels, laser_channels, time_value_list, cal_scanspeed, cal_points, cal_start_pos = extract_parameters(content)

        # Assume all selected files are from the same instrument and has the same polynomial coefficients
        if file_ind == 0:
            try:
                #  Read default polynomial file
                with open(default_poly_file, 'r') as p_file:
                    polycoeffs = [0] * channels
                    for p_ind in range(channels):
                        coeff = p_file.readline().split('\t')
                        polycoeffs[p_ind] = list(map(float, coeff))

                poly_path, poly_file_name = os.path.split(default_poly_file)

            except FileNotFoundError:
                print('Error: Polynomial coefficients not found in default location')

                poly_file = askopenfilename(title='Error! Select correct polynomial coefficients (.csv)')

                #  Read selected polynomial file
                with open(poly_file, 'r') as p_file:
                    polycoeffs = [0] * channels
                    for p_ind in range(channels):
                        coeff = p_file.readline().split('\t')
                        polycoeffs[p_ind] = list(map(float, coeff))

                poly_path, poly_file_name = os.path.split(poly_file)

            except ValueError:
                print('Error: Default polynomial coefficients not matching number of channels')

                poly_file = askopenfilename(title='Error! Select correct polynomial coefficients (.csv)')

                #  Read selected polynomial file
                with open(poly_file, 'r') as p_file:
                    polycoeffs = [0] * channels
                    for p_ind in range(channels):
                        coeff = p_file.readline().split('\t')
                        polycoeffs[p_ind] = list(map(float, coeff))

                poly_path, poly_file_name = os.path.split(poly_file)

        #  Extract and calibrate spectra for each laser using parallel computing
        jobs = []
        process_step = 0
        while process_step < min(channels, logical_cores):
            pr = mp.Process(target=extract_spectra, args=(content, process_step, polycoeffs[process_step], start_pos, scan_speed, cal_scanspeed, cal_points, cal_start_pos, time_value_list, laser_channels, spr2_file))
            jobs.append(pr)
            pr.start()
            print('Started working on channel L' + str(process_step + 1) + ' ' + str(laser_channels[process_step]) + 'nm')
            process_step += 1

        # If there are fewer logical cores than channels, wait for the first few jobs to finish, then start new processes as they do
        remaining_process_step = 0
        while process_step < channels:
            jobs[remaining_process_step].join()
            pr = mp.Process(target=extract_spectra, args=(content, process_step, polycoeffs[process_step], start_pos, scan_speed, cal_scanspeed, cal_points, cal_start_pos, time_value_list, laser_channels, spr2_file))
            jobs.append(pr)
            pr.start()
            print('Started working on channel L' + str(process_step + 1) + ' ' + str(laser_channels[process_step]) + 'nm')
            process_step += 1
            remaining_process_step += 1

        # Wait for the jobs of the first file to finish
        for job in jobs:
            job.join()

        print('File: ' + spr2_file_name + ' is done.')





import os
import numpy as np
from datetime import datetime
import SPRpy_functions
import SPRpy_spr2_to_csv
# Run this script to get polynomial coefficients translating motor step values to angles (needed for spr2_to_csv.py)
# Cannot take angle values directly as they may differ slightly in number of points between instrument reboots (step motor homing sequence)
# Polyfit "over fitting" on steps vs angles from .spr2 for each wavelength.
# This gives absolute angle value accuracy within the noise limit of the homing sequence.
# Use spectra from full angular scans
# Check from time to time that the angles obtained from spr2_to_csv conversion matches the angles from .dto export for a given scan.
# Small deviations of < 10-20 mdeg in absolute angle is likely within normal instrument uncertainty.
# If the step motor has been significantly affected or worn, it could be time to run this script again and replace this file for the spr2_to_csv.py script

# Prompt for required files (they are sorted by default)
spr2_file = SPRpy_functions.select_file('Select .spr2 file of full range slow speed angular scan.', file_types=[('SPR2 file', '*.spr2')])
dto_files = SPRpy_functions.select_files('Select .dto Bionavs viewer exported files for all wavelengths of full range slow speed angular scan.', file_types=[('DTO file', '*.dto')])
dto_wavelengths = [int(dto_file[-9:-6]) for dto_file in dto_files]

# Read .spr2 file and get calibration values
with open(spr2_file, 'r') as file:
    spr2_content = file.read()
start_pos, scanspeed, _, _, _, _, _, _ = SPRpy_spr2_to_csv.extract_parameters(spr2_content)

# Get polynomial coefficients
polycoffs = []
for dto_file in dto_files:
    data = np.loadtxt(dto_file, delimiter='\t', converters=lambda s: s.replace(',', '.')).T
    angles = data[0, :]
    step_max = start_pos + scanspeed * len(angles)
    steps = np.arange(start_pos, step_max, scanspeed)
    p = np.polyfit(steps, angles, 6)
    polycoffs.append(p)
    y_fit = np.polyval(p, steps)
    Chi_squared = np.sum(((angles - y_fit) / np.std(angles)) ** 2)
    print(Chi_squared)

# Write result to .csv file
polycoff_matrix = np.vstack(tuple(polycoffs))
cal_save_folder = os.getcwd()
file_name = f"{cal_save_folder}/SPRpy_X_cal_values_{datetime.now().strftime('%y-%m-%d')}.csv"
np.savetxt(file_name, polycoff_matrix, delimiter='\t')




import platform
# Check if OS is Windows
if platform.system() != 'Windows':
    raise OSError('This script is only for Windows OS')

import sysconfig
from win32com.client import Dispatch
import os


def create_shortcut():

    # folder name
    folder_name = 'SPRpy'

    # Scripts directory (location of python packages)
    lib_dir = sysconfig.get_path('purelib')

    # Target of shortcut
    target = os.path.join(lib_dir, folder_name)

    # Name of link file
    link_name = folder_name + '.lnk'

    # Path to location of link file
    shell = Dispatch('WScript.Shell')
    shell_folder = shell.SpecialFolders('Desktop')
    shortcut = shell.CreateShortCut(shell_folder + '\\' + link_name)
    shortcut.TargetPath = target
    shortcut.WorkingDirectory = lib_dir
    shortcut.save()



# Contains fresnel functions for the transfer-matrix method and curve fitting and TIR angle determination

import numpy as np
import pandas as pd
import bottleneck


def fresnel_calculation(fitted_var=None,
                        fitted_layer_index=(2, 3),
                        angles=np.linspace(39, 50, 1567),
                        wavelength=670,
                        layer_thicknesses=np.array([np.nan, 2, 50, 4, np.nan]),
                        n_re=np.array([1.5202, 3.3105, 0.2238, 1.5, 1.0003]),
                        n_im=np.array([0, 3.4556, 3.9259, 0, 0]),
                        ydata=None,
                        ydata_type='R',
                        weights=None,
                        polarization=1.0,
                        ydata_offset=0,
                        ):

    """
    Function for calculating fresnel coefficients or for fitting angular reflectivity traces based on the residuals of
    a measurement. By default, the function provides the thickness of a monolayer of BSA on gold in air.

    :param fitted_var: variable to be fitted
    :param angles: ndarray
    :param fitted_layer_index: tuple
    :param wavelength: int
    :param layer_thicknesses: ndarray
    :param n_re: ndarray
    :param n_im: ndarray
    :param ydata: ndarray (default None), if provided the function will instead return residuals between the modelled intensity and measurement
    :param ydata_type: string, specify if reflectivity ('R'), transmission ('T') or absorption ('A') is fitted against
    :param polarization: int, 1 (default) or 0
    :return: ndarray(s), either the fresnel coefficients or the residuals between modelled intensity and measured intensity
    """

    # Check first if fitting is performed or not
    if fitted_var is not None:

        # Selecting main layer to fit
        match fitted_layer_index[1]:
            case 0:
                print('Invalid fitting variable!')
                return 0
            case 1:
                layer_thicknesses[fitted_layer_index[0]] = fitted_var[0]
            case 2:
                n_re[fitted_layer_index[0]] = fitted_var[0]
            case 3:
                n_im[fitted_layer_index[0]] = fitted_var[0]

        # Include fitting intensity offset
        if len(fitted_var) >= 2:
            ydata_offset = fitted_var[1]

        # Include fitting prism extinction value
        if len(fitted_var) == 3:
            n_im[0] = fitted_var[2]

    # Merge real and imaginary refractive indices
    n = n_re + 1j * n_im

    # Calculate fresnel coefficients for every angle
    fresnel_coefficients_reflection = np.zeros(len(angles))
    fresnel_coefficients_transmission = np.zeros(len(angles))
    fresnel_coefficients_absorption = np.zeros(len(angles))

    for angle_ind, angle_val in enumerate(angles):

        # Snell's law
        theta = np.zeros(len(n), dtype=np.complex128)
        theta[0] = angle_val * np.pi / 180
        for a in range(len(n) - 1):
            theta[a + 1] = np.real(np.arcsin(n[a] / n[a + 1] * np.sin(theta[a]))) - 1j * np.abs(np.imag(np.arcsin(n[a] / n[a + 1] * np.sin(theta[a]))))

        # Calculating fresnel coefficients:
        fresnel_reflection = np.zeros(len(n) - 1, dtype=np.complex128)
        fresnel_transmission = np.zeros(len(n) - 1, dtype=np.complex128)

        for a in range(len(n) - 1):
            # formulas for s polarization
            s_reflection = (n[a] * np.cos(theta[a]) - n[a + 1] * np.cos(theta[a + 1])) / (n[a] * np.cos(theta[a]) + n[a + 1] * np.cos(theta[a + 1]))
            s_transmission = 2 * n[a] * np.cos(theta[a]) / (n[a] * np.cos(theta[a]) + n[a + 1] * np.cos(theta[a + 1]))

            # formulas for p polarization
            p_reflection = (n[a] * np.cos(theta[a + 1]) - n[a + 1] * np.cos(theta[a])) / (n[a] * np.cos(theta[a + 1]) + n[a + 1] * np.cos(theta[a]))
            p_transmission = 2 * n[a] * np.cos(theta[a]) / (n[a] * np.cos(theta[a + 1]) + n[a + 1] * np.cos(theta[a]))

            fresnel_reflection[a] = s_reflection*(polarization-1) + p_reflection*polarization
            fresnel_transmission[a] = s_transmission*(polarization-1) + p_transmission*polarization

        # Phase shift factors:
        delta = np.zeros(len(n) - 2, dtype=np.complex128)
        for a in range(len(n) - 2):
            delta[a] = 2 * np.pi * layer_thicknesses[a + 1] / wavelength * n[a + 1] * np.cos(theta[a + 1])

        # Build up transfer matrix:
        transfer_matrix = np.array([[1, 0], [0, 1]], dtype=np.complex128)
        for a in range(len(n) - 2):
            transfer_matrix = np.dot(transfer_matrix, np.dot(1 / fresnel_transmission[a], np.dot(np.array([[1, fresnel_reflection[a]], [fresnel_reflection[a], 1]]), np.array([[np.exp(-1j * delta[a]), 0], [0, np.exp(1j * delta[a])]]))))
        transfer_matrix = np.dot(transfer_matrix, np.dot(1 / fresnel_transmission[len(n) - 2], np.array([[1, fresnel_reflection[len(n) - 2]], [fresnel_reflection[len(n) - 2], 1]])))

        # Total fresnel coefficients:
        fr_tot = transfer_matrix[1, 0] / transfer_matrix[0, 0]
        ft_tot = 1 / transfer_matrix[0, 0]

        # Special case of single interface:
        if len(n) == 2:
            fr_tot = fresnel_reflection[0]
            ft_tot = fresnel_transmission[0]

        # Total fresnel coefficients in intensity:
        fresnel_coefficients_reflection[angle_ind] = np.absolute(fr_tot)**2
        fresnel_coefficients_transmission[angle_ind] = np.absolute(ft_tot)**2 * np.real(n[-1] * np.cos(theta[-1])) / np.real(n[0] * np.cos(theta[0]))
        fresnel_coefficients_absorption[angle_ind] = 1 - fresnel_coefficients_reflection[angle_ind] - fresnel_coefficients_transmission[angle_ind]

    # Return fresnel coefficients or residuals depending on if fitting is performed against ydata
    if ydata is None:
        match ydata_type:
            case 'R':
                return fresnel_coefficients_reflection - ydata_offset
            case 'T':
                return fresnel_coefficients_transmission - ydata_offset
            case 'A':
                return fresnel_coefficients_absorption - ydata_offset

    else:
        fresnel_residuals = np.array([]*len(ydata))
        match ydata_type:
            case 'R':
                fresnel_residuals = (fresnel_coefficients_reflection - ydata_offset) - ydata
            case 'T':
                fresnel_residuals = (fresnel_coefficients_transmission - ydata_offset) - ydata
            case 'A':
                fresnel_residuals = (fresnel_coefficients_absorption - ydata_offset) - ydata
        if weights is not None:
            return fresnel_residuals*weights
        else:
            return fresnel_residuals


def TIR_determination(xdata, ydata, SPR_TIR_fitting_parameters):

    # Convert to numpy array first if necessary
    if isinstance(xdata, pd.Series):
        xdata = xdata.to_numpy()

    if isinstance(ydata, pd.Series):
        ydata = ydata.to_numpy()

    TIR_ydata = ydata[(xdata >= SPR_TIR_fitting_parameters['TIR range'][0]) & (xdata <= SPR_TIR_fitting_parameters['TIR range'][1])]
    TIR_xdata = xdata[(xdata >= SPR_TIR_fitting_parameters['TIR range'][0]) & (xdata <= SPR_TIR_fitting_parameters['TIR range'][1])]

    # Filter the data with a moving-average filter to smoothen the signal
    TIR_ydata_filtered = bottleneck.move_mean(TIR_ydata, window=SPR_TIR_fitting_parameters['TIR window count'], min_count=1)
    TIR_xdata_filtered = bottleneck.move_mean(TIR_xdata, window=SPR_TIR_fitting_parameters['TIR window count'], min_count=1)

    # Find maximum derivative
    deriv_ydata = np.concatenate(([np.diff(TIR_ydata_filtered)[0]], np.diff(TIR_ydata_filtered)))  # Add extra value for dimensions
    dTIR_i = np.argmax(deriv_ydata)

    # Fit against the derivative spike where the derivative is max, considering also nearest neighbors
    poly_c = np.polyfit(TIR_xdata_filtered[dTIR_i - SPR_TIR_fitting_parameters['points_below_TIR_peak']:dTIR_i + SPR_TIR_fitting_parameters['points_above_TIR_peak'] + 1],
                        deriv_ydata[dTIR_i - SPR_TIR_fitting_parameters['points_below_TIR_peak']:dTIR_i + SPR_TIR_fitting_parameters['points_above_TIR_peak'] + 1], 3)

    # Recreate the curve with a lot more points
    deriv_TIR_fit_x = np.linspace(TIR_xdata_filtered[dTIR_i - SPR_TIR_fitting_parameters['points_below_TIR_peak']], TIR_xdata_filtered[dTIR_i + SPR_TIR_fitting_parameters['points_above_TIR_peak']], SPR_TIR_fitting_parameters['TIR fit points'])

    # Find TIR from max of deriv fit
    deriv_TIR_fit_y = np.polyval(poly_c, deriv_TIR_fit_x)
    dTIR_final = np.argmax(deriv_TIR_fit_y)
    TIR_theta = deriv_TIR_fit_x[dTIR_final]

    return TIR_theta, TIR_xdata_filtered, deriv_ydata, deriv_TIR_fit_x, deriv_TIR_fit_y



# This file contains all SPRpy classes, and also methods and functions that depend on the class objects

import datetime
import os
import scipy
import pickle
import copy
import bottleneck
import multiprocessing
from SPRpy_functions import *
from fresnel_transfer_matrix import fresnel_calculation


class Session:

    """
    A base class for storing and loading a group of measurements for a particular session in one file. This should be
    the first thing that a user is prompted for before they start their analysis.
    """

    def __init__(self, version, SPR_TIR_fitting_parameters, name='Session', directory=None, current_data_path=None):
        self.version = version
        self.name = datetime.datetime.now().__str__()[0:16].replace(':', '_') + ' ' + name
        if not directory:
            directory = os.getcwd().replace('\\', '/')
            if os.path.exists(directory + '/SPRpy sessions'):
                self.location = directory + '/SPRpy sessions' + '/{name_}'.format(name_=self.name)
            else:
                os.mkdir(directory + '/SPRpy sessions')
                self.location = directory + '/SPRpy sessions' + '/{name_}'.format(name_=self.name)
        else:
            if os.path.exists(directory):
                self.location = directory + '/{name_}'.format(name_=self.name)
            else:
                directory = os.getcwd().replace('\\', '/')
                if os.path.exists(directory + '/SPRpy sessions'):
                    self.location = directory + '/SPRpy sessions' + '/{name_}'.format(name_=self.name)
                else:
                    os.mkdir(directory + '/SPRpy sessions')
                    self.location = directory + '/SPRpy sessions' + '/{name_}'.format(name_=self.name)
                print('Warning! Custom session folder in config.toml does not exist. Sessions placed in default SPRpy location instead.')

        if not os.path.exists(self.location):
            os.mkdir(self.location)
        if not os.path.exists(self.location + '/Sensors'):
            os.mkdir(self.location + '/Sensors')
        if not os.path.exists(self.location + '/Analysis instances'):
            os.mkdir(self.location + '/Analysis instances')
        self.sensor_instances = {}  # NOTE: The sessions in this list are also updated when modified as current sensor object
        self.sensor_ID_count = 0
        self.fresnel_analysis_instances = {}
        self.fresnel_analysis_ID_count = 0
        self.exclusion_height_analysis_instances = {}
        self.exclusion_height_analysis_ID_count = 0
        self.current_data_path = current_data_path
        self.SPR_TIR_fitting_parameters = SPR_TIR_fitting_parameters
        self.log = datetime.datetime.now().__str__()[0:16] + ' >> ' + 'Welcome to SPRpy!' \
            + '\n' + datetime.datetime.now().__str__()[0:16] + ' >> ' + 'Start your session by defining your SPR sensor layers.'

    def update_name_and_location(self, new_name):
        """
        Updates the name and location of the session.
        :param new_name: string
        :return:
        """

        self.name = new_name
        # old_location = self.location
        os.rename(self.location, self.location.replace(self.location.split('/')[-1], new_name))
        self.location = self.location.replace(self.location.split('/')[-1], new_name)
        return

    def remove_sensor(self, sensor_object_id):
        """
        Remove a sensor object from the session.
        :return:
        """
        removed = self.sensor_instances.pop(sensor_object_id)
        removed_file_path = self.location + '/Sensors' + '/S{id} {name}.pickle'.format(id=removed.object_id, name=removed.name)
        os.remove(removed_file_path)
        print('Removed the following sensor object: S{id} {name}'.format(id=removed.object_id, name=removed.name))

        return

    def remove_fresnel_analysis(self, analysis_object_id):
        """
        Remove an analysis object from the session.
        :return:
        """
        removed = self.fresnel_analysis_instances.pop(analysis_object_id)
        removed_file_path = self.location + '/Analysis instances' + '/FM{id} {name}.pickle'.format(id=removed.object_id, name=removed.name)
        os.remove(removed_file_path)
        print('Removed the following analysis object: FM{id} {name}'.format(id=removed.object_id, name=removed.name))

        return

    def remove_exclusion_height_analysis(self, analysis_object_id):
        """
        Remove an analysis object from the session.
        :return:
        """
        removed = self.exclusion_height_analysis_instances.pop(analysis_object_id)
        removed_file_path = self.location + '/Analysis instances' + '/EH{id} {name}.pickle'.format(id=removed.object_id, name=removed.name)
        os.remove(removed_file_path)
        print('Removed the following analysis object: EH{id} {name}'.format(id=removed.object_id, name=removed.name))

        return

    def save_all(self):
        """
        Saves all objects stored in the session, and the session file itself.
        :return: None
        """

        # Save session object
        with open(self.location + '/Session file (v{version_}).pickle'.format(version_=self.version.replace('.', '_')), 'wb') as save_file:
            pickle.dump(self, save_file)

        # Save sensor instances
        for sensor_id in self.sensor_instances:
            with open(self.location + '/Sensors' + '/S{id} {name}.pickle'.format(id=sensor_id, name=self.sensor_instances[sensor_id].name), 'wb') as save_file:
                pickle.dump(self.sensor_instances[sensor_id], save_file)

        # Save fresnel analysis instances
        for analysis_id in self.fresnel_analysis_instances:
            with open(self.location + '/Analysis instances' + '/FM{id} {name}.pickle'.format(id=analysis_id, name=self.fresnel_analysis_instances[analysis_id].name), 'wb') as save_file:
                pickle.dump(self.fresnel_analysis_instances[analysis_id], save_file)

        # Save exclusion height analysis instances
        for analysis_id in self.exclusion_height_analysis_instances:
            with open(self.location + '/Analysis instances' + '/FM{id} {name}.pickle'.format(id=analysis_id, name=self.exclusion_height_analysis_instances[analysis_id].name), 'wb') as save_file:
                pickle.dump(self.exclusion_height_analysis_instances[analysis_id], save_file)

        return

    def save_session(self):

        # Save session object
        with open(self.location + '/Session file (v{version_}).pickle'.format(version_=self.version.replace('.', '_')), 'wb') as save_file:
            pickle.dump(self, save_file)

        return

    def save_sensor(self, sensor_id):
        """
        Saves a single sensor object to the session.
        :return: None
        """

        with open(self.location + '/Sensors' + '/S{id} {name}.pickle'.format(id=sensor_id, name=self.sensor_instances[
            sensor_id].name), 'wb') as save_file:
            pickle.dump(self.sensor_instances[sensor_id], save_file)

        return

    def save_fresnel_analysis(self, analysis_id):
        """
        Saves a single fresnel analysis object to the session.
        :return: None
        """

        with open(self.location + '/Analysis instances' + '/FM{id} {name}.pickle'.format(id=analysis_id, name=self.fresnel_analysis_instances[analysis_id].name), 'wb') as save_file:
            pickle.dump(self.fresnel_analysis_instances[analysis_id], save_file)

        return

    def save_exclusion_height_analysis(self, analysis_id):
        """
        Saves a single fresnel analysis object to the session.
        :return: None
        """

        with open(self.location + '/Analysis instances' + '/EH{id} {name}.pickle'.format(id=analysis_id, name=self.exclusion_height_analysis_instances[analysis_id].name), 'wb') as save_file:
            pickle.dump(self.exclusion_height_analysis_instances[analysis_id], save_file)

        return

    def import_sensor(self):

        file_path_ = select_file(prompt='Select the sensor object', prompt_folder=self.location + '/Sensors')
        self.sensor_ID_count += 1

        with open(file_path_, 'rb') as import_file:
            sensor_object = pickle.load(import_file)

        sensor_object.object_id = self.sensor_ID_count
        self.sensor_instances[self.sensor_ID_count] = sensor_object

        return

    def import_fresnel_analysis(self):
        file_path_ = select_file(prompt='Select the analysis object', prompt_folder=self.location + '/Analysis instances')
        self.fresnel_analysis_ID_count += 1

        with open(file_path_, 'rb') as import_file:
            analysis_object = pickle.load(import_file)

        analysis_object.object_id = self.fresnel_analysis_ID_count
        self.fresnel_analysis_instances[analysis_object.object_id] = analysis_object

        return

    def import_exclusion_height_analysis(self):
        file_path_ = select_file(prompt='Select the analysis object', prompt_folder=self.location + '/Analysis instances')
        self.exclusion_height_analysis_ID_count += 1

        with open(file_path_, 'rb') as import_file:
            analysis_object = pickle.load(import_file)

        analysis_object.object_id = self.exclusion_height_analysis_ID_count
        self.exclusion_height_analysis_instances[analysis_object.object_id] = analysis_object

        return

class Sensor:

    """
      An SPR measurement typically have some things in common, such as the sensor layers, measured angles,
    measured reflectivity, measurement time, etc. This information can be shared between different analysis methods for
    one measurement. This class serves as a basis for describing the current sensor, containing information about its
    layers and their optical properties.
    """

    def __init__(self, data_path_, object_id_, default_sensor_values, object_name_='Gold sensor', sensor_metal='Au', data_type='R'):

        # Load sensor's default optical properties
        self.object_id = object_id_
        self.name = object_name_
        self.data_path = data_path_
        self.data_type = data_type
        self.wavelength = int(data_path_[-9:-6])
        self.channel = data_path_[-12:-4].replace('_', ' ')
        self.sensor_metal = sensor_metal
        self.default_sensor_values = default_sensor_values
        self.set_default_optical_properties(self.sensor_metal)
        self.fitted_var = self.optical_parameters.iloc[self.fitted_layer_index]

    def set_default_optical_properties(self, sensor_metal):

        # These default parameters should be set based on material layer and wavelength from loaded .csv file

        match sensor_metal:
            case 'Au' | 'gold' | 'Gold' | 'GOLD':
                self.layer_thicknesses = np.array([np.nan, float(self.default_sensor_values["d_nm"][1]), float(self.default_sensor_values["d_nm"][2]), np.nan])
                self.fitted_layer_index = (2, 3)  # Tuple with index for df.iloc[fitted_layer_index]
                match self.wavelength:
                    case 670:
                        self.refractive_indices = np.array([float(self.default_sensor_values["n_670"][0]), float(self.default_sensor_values["n_670"][1]), float(self.default_sensor_values["n_670"][2]), 1.0003])
                        self.extinction_coefficients = np.array([float(self.default_sensor_values["k_670"][0]), float(self.default_sensor_values["k_670"][1]), float(self.default_sensor_values["k_670"][2]), 0.0])
                    case 785:
                        self.refractive_indices = np.array([float(self.default_sensor_values["n_785"][0]), float(self.default_sensor_values["n_785"][1]), float(self.default_sensor_values["n_785"][2]), 1.0003])
                        self.extinction_coefficients = np.array([float(self.default_sensor_values["k_785"][0]), float(self.default_sensor_values["k_785"][1]), float(self.default_sensor_values["k_785"][2]), 0.0])
                    case 850:
                        self.refractive_indices = np.array([float(self.default_sensor_values["n_850"][0]), float(self.default_sensor_values["n_850"][1]), float(self.default_sensor_values["n_850"][2]), 1.0003])
                        self.extinction_coefficients = np.array([float(self.default_sensor_values["k_850"][0]), float(self.default_sensor_values["k_850"][1]), float(self.default_sensor_values["k_850"][2]), 0.0])
                    case 980:
                        self.refractive_indices = np.array([float(self.default_sensor_values["n_980"][0]), float(self.default_sensor_values["n_980"][1]), float(self.default_sensor_values["n_980"][2]), 1.0003])
                        self.extinction_coefficients = np.array([float(self.default_sensor_values["k_980"][0]), float(self.default_sensor_values["k_980"][1]), float(self.default_sensor_values["k_980"][2]), 0.0])
                self.optical_parameters = pd.DataFrame(data={'Layers': ['Prism', 'Cr', 'Au', 'Bulk'],  # NOTE: sensor_object.optical_parameters will auto-update when changing sensor_object.layer_thicknesses, sensor_object.refractive_indices or sensor_object.extinction_coefficients
                                                             'd [nm]': self.layer_thicknesses,
                                                             'n': self.refractive_indices,
                                                             'k': self.extinction_coefficients})

            case 'sio2' | 'SiO2' | 'SIO2' | 'Glass' | 'glass' | 'silica':
                # Fused silica values source: L. V. Rodríguez-de Marcos, J. I. Larruquert, J. A. Méndez, J. A. Aznárez. Self-consistent optical constants of SiO2 and Ta2O5 films. Opt. Mater. Express 6, 3622-3637 (2016)
                self.layer_thicknesses = np.array([np.nan, float(self.default_sensor_values["d_nm"][1]), float(self.default_sensor_values["d_nm"][2]),  float(self.default_sensor_values["d_nm"][3]), np.nan])
                self.fitted_layer_index = (3, 1)  # Tuple with index for df.iloc[fitted_layer_index]
                match self.wavelength:
                    case 670:
                        self.refractive_indices = np.array([float(self.default_sensor_values["n_670"][0]), float(self.default_sensor_values["n_670"][1]), float(self.default_sensor_values["n_670"][2]), float(self.default_sensor_values["n_670"][3]), 1.0003])
                        self.extinction_coefficients = np.array([float(self.default_sensor_values["k_670"][0]), float(self.default_sensor_values["k_670"][1]), float(self.default_sensor_values["k_670"][2]), float(self.default_sensor_values["k_670"][3]), 0.0])
                    case 785:
                        self.refractive_indices = np.array([float(self.default_sensor_values["n_785"][0]), float(self.default_sensor_values["n_785"][1]), float(self.default_sensor_values["n_785"][2]), float(self.default_sensor_values["n_785"][3]), 1.0003])
                        self.extinction_coefficients = np.array([float(self.default_sensor_values["k_785"][0]), float(self.default_sensor_values["k_785"][1]), float(self.default_sensor_values["k_785"][2]), float(self.default_sensor_values["k_785"][3]), 0.0])
                    case 850:
                        self.refractive_indices = np.array([float(self.default_sensor_values["n_850"][0]), float(self.default_sensor_values["n_850"][1]), float(self.default_sensor_values["n_850"][2]), float(self.default_sensor_values["n_850"][3]), 1.0003])
                        self.extinction_coefficients = np.array([float(self.default_sensor_values["k_850"][0]), float(self.default_sensor_values["k_850"][1]), float(self.default_sensor_values["k_850"][2]), float(self.default_sensor_values["k_850"][3]), 0.0])
                    case 980:
                        self.refractive_indices = np.array([float(self.default_sensor_values["n_980"][0]), float(self.default_sensor_values["n_980"][1]), float(self.default_sensor_values["n_980"][2]), float(self.default_sensor_values["n_980"][3]), 1.0003])
                        self.extinction_coefficients = np.array([float(self.default_sensor_values["k_980"][0]), float(self.default_sensor_values["k_980"][1]), float(self.default_sensor_values["k_980"][2]), float(self.default_sensor_values["k_980"][3]), 0.0])
                self.optical_parameters = pd.DataFrame(data={'Layers': ['Prism', 'Cr', 'Au', 'SiO2', 'Bulk'],
                                                             'd [nm]': self.layer_thicknesses,
                                                             'n': self.refractive_indices,
                                                             'k': self.extinction_coefficients})

            case 'Pd' | 'palladium' | 'Palladium' | 'PALLADIUM':
                # Source Pd 670nm : Andersson, John, et al. "Surface plasmon resonance sensing with thin films of palladium and platinum–quantitative and real-time analysis." Physical Chemistry Chemical Physics 24.7 (2022): 4588-4594.
                # Source Pd 785nm, 980nm: Similar procedure as Andersson, John et al., but these were not reported.
                self.layer_thicknesses = np.array([np.nan, float(self.default_sensor_values["d_nm"][1]), float(self.default_sensor_values["d_nm"][4]), np.nan])
                self.fitted_layer_index = (2, 3)  # Tuple with index for df.iloc[fitted_layer_index]
                match self.wavelength:
                    case 670:
                        self.refractive_indices = np.array(
                            [float(self.default_sensor_values["n_670"][0]), float(self.default_sensor_values["n_670"][1]), float(self.default_sensor_values["n_670"][4]), 1.0003])
                        self.extinction_coefficients = np.array(
                            [float(self.default_sensor_values["k_670"][0]), float(self.default_sensor_values["k_670"][1]), float(self.default_sensor_values["k_670"][4]), 0.0])
                    case 785:
                        self.refractive_indices = np.array(
                            [float(self.default_sensor_values["n_785"][0]), float(self.default_sensor_values["n_785"][1]), float(self.default_sensor_values["n_785"][4]), 1.0003])
                        self.extinction_coefficients = np.array(
                            [float(self.default_sensor_values["k_785"][0]), float(self.default_sensor_values["k_785"][1]), float(self.default_sensor_values["k_785"][4]), 0.0])
                    case 850:
                        self.refractive_indices = np.array(
                            [float(self.default_sensor_values["n_850"][0]), float(self.default_sensor_values["n_850"][1]), float(self.default_sensor_values["n_850"][4]), 1.0003])
                        self.extinction_coefficients = np.array(
                            [float(self.default_sensor_values["k_850"][0]), float(self.default_sensor_values["k_850"][1]), float(self.default_sensor_values["k_850"][4]), 0.0])
                    case 980:
                        self.refractive_indices = np.array(
                            [float(self.default_sensor_values["n_980"][0]), float(self.default_sensor_values["n_980"][1]), float(self.default_sensor_values["n_980"][4]), 1.0003])
                        self.extinction_coefficients = np.array(
                            [float(self.default_sensor_values["k_980"][0]), float(self.default_sensor_values["k_980"][1]), float(self.default_sensor_values["k_980"][4]), 0.0])
                self.optical_parameters = pd.DataFrame(data={'Layers': ['Prism', 'Cr', 'Pd', 'Bulk'],
                                                             'd [nm]': self.layer_thicknesses,
                                                             'n': self.refractive_indices,
                                                             'k': self.extinction_coefficients})

            case 'Pt' | 'platinum' | 'Platinum' | 'PLATINUM':
                # Source Pt 670 nm: Andersson, John, et al. "Surface plasmon resonance sensing with thin films of palladium and platinum–quantitative and real-time analysis." Physical Chemistry Chemical Physics 24.7 (2022): 4588-4594.
                # Source Pt 785, 850 and 980 nm: W. S. M. Werner, K. Glantschnig, C. Ambrosch-Draxl. Optical constants and inelastic electron-scattering data for 17 elemental metals, J. Phys Chem Ref. Data 38, 1013-1092 (2009)
                self.layer_thicknesses = np.array([np.nan, 2.00, 20.00, np.nan])
                self.fitted_layer_index = (2, 3)  # Tuple with index for df.iloc[fitted_layer_index]
                match self.wavelength:
                    case 670:
                        self.refractive_indices = np.array(
                            [float(self.default_sensor_values["n_670"][0]), float(self.default_sensor_values["n_670"][1]), float(self.default_sensor_values["n_670"][5]), 1.0003])
                        self.extinction_coefficients = np.array(
                            [float(self.default_sensor_values["k_670"][0]), float(self.default_sensor_values["k_670"][1]), float(self.default_sensor_values["k_670"][5]), 0.0])
                    case 785:
                        self.refractive_indices = np.array(
                            [float(self.default_sensor_values["n_785"][0]), float(self.default_sensor_values["n_785"][1]), float(self.default_sensor_values["n_785"][5]), 1.0003])
                        self.extinction_coefficients = np.array(
                            [float(self.default_sensor_values["k_785"][0]), float(self.default_sensor_values["k_785"][1]), float(self.default_sensor_values["k_785"][5]), 0.0])
                    case 850:
                        self.refractive_indices = np.array(
                            [float(self.default_sensor_values["n_850"][0]), float(self.default_sensor_values["n_850"][1]), float(self.default_sensor_values["n_850"][5]), 1.0003])
                        self.extinction_coefficients = np.array(
                            [float(self.default_sensor_values["k_850"][0]), float(self.default_sensor_values["k_850"][1]), float(self.default_sensor_values["k_850"][5]), 0.0])
                    case 980:
                        self.refractive_indices = np.array(
                            [float(self.default_sensor_values["n_980"][0]), float(self.default_sensor_values["n_980"][1]), float(self.default_sensor_values["n_980"][5]), 1.0003])
                        self.extinction_coefficients = np.array(
                            [float(self.default_sensor_values["k_980"][0]), float(self.default_sensor_values["k_980"][1]), float(self.default_sensor_values["k_980"][5]), 0.0])
                self.optical_parameters = pd.DataFrame(data={'Layers': ['Prism', 'Cr', 'Pt', 'Bulk'],
                                                             'd [nm]': self.layer_thicknesses,
                                                             'n': self.refractive_indices,
                                                             'k': self.extinction_coefficients})
        return

class FresnelModel:
    """
    This class defines how a modelled reflectivity trace behaves.

    TODO: Each object should also have a .csv export function.

    """

    def __init__(self, session_object, sensor_object_, data_path_, reflectivity_df_, object_id_, object_name_):
        self.name = object_name_
        self.object_id = object_id_
        self.sensor_object = sensor_object_
        self.sensor_object_label = ''
        self.initial_data_path = data_path_
        self.measurement_data = reflectivity_df_
        self.SPR_TIR_fitting_parameters = session_object.SPR_TIR_fitting_parameters  # We want the SPR and TIR fitting parameters to update when they change globally in the session
        self.angle_range = session_object.SPR_TIR_fitting_parameters['Fresnel_angle_range_points']
        self.polarization = 1.0  # 0-1, for degree of s (=0) and p(=1) polarization,
        self.ini_guess = np.array(4)
        self.bounds = [0, 50]  # or [(lb1, lb2), (ub1, ub2)] etc
        self.extinction_correction = 0
        self.y_offset = 0
        self.fit_offset = True
        self.fit_prism_k = True
        self.fitted_data = None
        self.fitted_result = None
        self.fitted_layer_index = copy.deepcopy(self.sensor_object.fitted_layer_index)  # Deepcopy to not overwrite if fitting multiple layers in the same sensor object
        self.fitted_layer = copy.deepcopy(self.sensor_object.optical_parameters.iloc[self.fitted_layer_index[0], 0])

    def calculate_fresnel_trace(self):

        fresnel_coefficients_ = fresnel_calculation(None,
                                                    angles=self.angle_range,
                                                    fitted_layer_index=self.sensor_object.fitted_layer_index,
                                                    wavelength=self.sensor_object.wavelength,
                                                    layer_thicknesses=self.sensor_object.layer_thicknesses,
                                                    n_re=self.sensor_object.refractive_indices,
                                                    n_im=self.sensor_object.extinction_coefficients,
                                                    ydata=None,
                                                    ydata_type=self.sensor_object.data_type,
                                                    polarization=self.polarization)
        return fresnel_coefficients_

    def model_reflectivity_trace(self):

        xdata_ = self.measurement_data['angles']
        ydata_ = self.measurement_data['ydata']

        # Calculate TIR angle and bulk refractive index
        TIR_angle, _, _, _, _ = TIR_determination(xdata_, ydata_, self.SPR_TIR_fitting_parameters)
        self.sensor_object.refractive_indices[-1] = self.sensor_object.refractive_indices[0] * np.sin(np.pi / 180 * TIR_angle)

        # Add extinction correction to fitted surface layer extinction value
        self.sensor_object.extinction_coefficients[0] += self.extinction_correction  # Manually correct prism layer extinction

        # Selecting a range of measurement data to use for fitting, and including an offset in reflectivity (iterated 3 times)
        selection_xdata_ = xdata_[(xdata_ >= self.angle_range[0]) & (xdata_ <= self.angle_range[1])]
        selection_ydata_ = ydata_[(xdata_ >= self.angle_range[0]) & (xdata_ <= self.angle_range[1])]

        # Weighing options
        weights = None

        # weights_ = np.abs(np.diff(selection_ydata_))+1  # Highest derivative
        # weights = np.append(weights_, 0)

        # weights = 1/selection_ydata_

        # weights = np.ones(len(selection_ydata_))
        # weights[selection_ydata_.argmin():-1] = 2

        # Perform the first fitting
        result = scipy.optimize.least_squares(fresnel_calculation,
                                              self.ini_guess,
                                              bounds=self.bounds,
                                              kwargs={'fitted_layer_index': self.sensor_object.fitted_layer_index,
                                                      'wavelength': self.sensor_object.wavelength,
                                                      'layer_thicknesses': self.sensor_object.layer_thicknesses,
                                                      'n_re': self.sensor_object.refractive_indices,
                                                      'n_im': self.sensor_object.extinction_coefficients,
                                                      'angles': selection_xdata_,
                                                      'ydata': selection_ydata_,
                                                      'weights': weights,
                                                      'ydata_type': self.sensor_object.data_type,
                                                      'polarization': self.polarization,
                                                      'ydata_offset': self.y_offset},
                                              loss='huber',
                                              ftol=1e-12,
                                              xtol=1e-12,
                                              gtol=1e-12)

        # Collect the results from least_squares object and calculate corresponding fresnel coefficients
        self.fitted_result = np.array(result['x'])

        if self.fit_offset:
            self.y_offset = self.fitted_result[1]
        else:
            self.y_offset = 0

        fresnel_coefficients = fresnel_calculation(self.fitted_result,
                                                   fitted_layer_index=self.sensor_object.fitted_layer_index,
                                                   angles=selection_xdata_,
                                                   wavelength=self.sensor_object.wavelength,
                                                   layer_thicknesses=self.sensor_object.layer_thicknesses,
                                                   n_re=self.sensor_object.refractive_indices,
                                                   n_im=self.sensor_object.extinction_coefficients,
                                                   ydata=None,
                                                   weights=weights,
                                                   ydata_type='R',
                                                   polarization=self.polarization,
                                                   ydata_offset=self.y_offset
                                                   )

        # Compile into fresnel_coefficients data frame
        self.fitted_data = pd.DataFrame(data={'angles': selection_xdata_, 'ydata': fresnel_coefficients})

        return self.fitted_data

    def export_fitted_results(self):
        """
        Exporting the result (including parameters) of a particular analysis as a .csv file
        :return:
        """
        pass

    def export_calculated_results(self):
        """
        Exporting the calculated fresnel traces (including parameters) of a particular analysis as a .csv file
        :return:
        """
        pass


class ExclusionHeight:

    """
        This class defines an analysis object for determining the exclusion height from measurement data with probe
        injections. The underlying method is described as the "non-interacting probe method" in the literature.
    """

    def __init__(self, session_object, fresnel_object_, sensorgram_df_, data_path_,  object_id_, object_name_):
        self.name = object_name_
        self.object_id = object_id_
        self.fresnel_object = fresnel_object_
        self.SPR_TIR_fitting_parameters = session_object.SPR_TIR_fitting_parameters  # We want the SPR and TIR fitting parameters to update when they change globally in the session
        self.fresnel_object_label = 'Fresnel background: FM{analysis_number} {analysis_name}'.format(
                                                                    analysis_number=fresnel_object_.object_id,
                                                                    analysis_name=fresnel_object_.name)
        self.sensor_object = fresnel_object_.sensor_object
        self.polarization = fresnel_object_.polarization
        self.initial_data_path = data_path_
        self.sensorgram_data = sensorgram_df_.iloc[:,:3]
        self.sensorgram_offset_ind = 0
        self.d_n_pair_resolution = 200
        self.height_steps = np.linspace(0, 200, self.d_n_pair_resolution)
        self.points_below_SPR_min_ind = None  # This will be calculated based on the most updated fits of the Fresnel background
        self.points_above_SPR_min_ind = None  # This will be calculated based on the most updated fits of the Fresnel background
        self.injection_points = []
        self.buffer_points = []
        self.probe_points = []
        self.SPR_vs_TIR_dfs = []  # List of dataframes with labels 'SPR angles' and 'TIR angles' for indexing each step result
        self.buffer_reflectivity_dfs = []  # Use labels 'buffer reflectivity' and 'buffer angles' (and likewise probe) for indexing
        self.buffer_bulk_RIs = []  # Calculated from TIR angle of each reflectivity DF
        self.probe_reflectivity_dfs = []  # Use labels 'buffer reflectivity' and 'buffer angles' (and likewise probe) for indexing
        self.probe_bulk_RIs = []  # Calculated from TIR angle of each reflectivity DF
        self.d_n_pair_dfs = []  # Use labels 'height' and 'buffer RI' and 'probe RI' for indexing
        self.all_exclusion_results = []
        self.mean_exclusion_height_result = None  # Tuple of mean value of exclusion height from all injection steps, and standard deviation
        self.mean_exclusion_RI_result = None  # Tuple of mean value of exclusion RI from all injection steps, and standard deviation
        self.abort_flag = False
        self.fit_offset = False
        self.fit_prism = False

    def initialize_model(self, ydata_df):

        # Calculate number of points above and below minimum point based on fresnel model background range
        background_reflectivity = self.fresnel_object.measurement_data['ydata']
        background_angles = self.fresnel_object.measurement_data['angles']
        selection_criterion = (background_angles >= self.fresnel_object.angle_range[0]) & (background_angles <= self.fresnel_object.angle_range[1])
        selection_ydata_series = background_reflectivity[selection_criterion]
        smoothened_selection = bottleneck.move_mean(selection_ydata_series.to_numpy(), window=4, min_count=1)  # Ensures closer fit to minimum position
        smoothened_selection_series = pd.Series(smoothened_selection)
        self.points_below_SPR_min_ind = len(smoothened_selection_series[(smoothened_selection_series.index < smoothened_selection_series.idxmin())])
        self.points_above_SPR_min_ind = len(smoothened_selection_series[(smoothened_selection_series.index > smoothened_selection_series.idxmin())])

        # Overwrite previous probe and buffer reflectivity data frames and bulk RI lists and SPR vs TIR data frames
        self.buffer_reflectivity_dfs = []
        self.buffer_bulk_RIs = []
        self.probe_reflectivity_dfs = []
        self.probe_bulk_RIs = []
        self.SPR_vs_TIR_dfs = []

        # Calculate average reflectivity traces based on selected points
        bufferpoint_index = 0
        for reflectivity_index in range(int(len(self.buffer_points) / 2)):

            sliced_buffer_reflectivity_spectras = ydata_df.iloc[self.buffer_points[bufferpoint_index][0]:self.buffer_points[bufferpoint_index + 1][0], :]  # Selecting all spectras between the pairwise selected buffer points
            mean_buffer_reflectivity = sliced_buffer_reflectivity_spectras.mean(axis=0).squeeze()

            # Calculate TIR and bulk RI for each mean spectra
            try:
                buffer_TIR_angle, _, _, _, _ = TIR_determination(background_angles.to_numpy(), mean_buffer_reflectivity.to_numpy(), self.SPR_TIR_fitting_parameters)
            except TypeError:
                raise TypeError('Something went wrong when selecting buffer points. Please clear and reselect them and avoid clicking other point markers.')
            self.buffer_bulk_RIs.append(self.sensor_object.refractive_indices[0] * np.sin(np.pi / 180 * buffer_TIR_angle))

            # Calculate appropriate range selection
            buffer_reflectivity_minimum_ind = pd.Series(bottleneck.move_mean(mean_buffer_reflectivity.to_numpy(), window=4, min_count=1)).idxmin()
            self.buffer_reflectivity_dfs.append(pd.DataFrame(data={'reflectivity': mean_buffer_reflectivity.iloc[buffer_reflectivity_minimum_ind - self.points_below_SPR_min_ind:buffer_reflectivity_minimum_ind + self.points_above_SPR_min_ind],
                                                                   'angles': background_angles.iloc[buffer_reflectivity_minimum_ind - self.points_below_SPR_min_ind:buffer_reflectivity_minimum_ind + self.points_above_SPR_min_ind]
                                                                   })
                                                )

            # Next pair of buffer point indices
            bufferpoint_index += 2

        probepoint_index = 0
        for reflectivity_index in range(int(len(self.probe_points) / 2)):
            sliced_probe_reflectivity_spectras = ydata_df.iloc[self.probe_points[probepoint_index][0]:self.probe_points[probepoint_index + 1][0], :]  # Selecting all spectras between the pairwise selected probe point
            mean_probe_reflectivity = sliced_probe_reflectivity_spectras.mean(axis=0).squeeze()

            # Calculate TIR and bulk RI for each mean spectra
            try:
                probe_TIR_angle, _, _, _, _ = TIR_determination(background_angles.to_numpy(), mean_probe_reflectivity.to_numpy(), self.SPR_TIR_fitting_parameters)
            except TypeError:
                raise TypeError('Something went wrong when selecting probe points. Please clear and reselect them and avoid clicking other point markers.')
            self.probe_bulk_RIs.append(self.sensor_object.refractive_indices[0] * np.sin(np.pi / 180 * probe_TIR_angle))

            # Calculate appropriate range selection
            probe_reflectivity_minimum_ind = pd.Series(bottleneck.move_mean(mean_probe_reflectivity.to_numpy(), window=4, min_count=1)).idxmin()
            self.probe_reflectivity_dfs.append(pd.DataFrame(data={'reflectivity': mean_probe_reflectivity.iloc[probe_reflectivity_minimum_ind - self.points_below_SPR_min_ind:probe_reflectivity_minimum_ind + self.points_above_SPR_min_ind],
                                                                  'angles': background_angles.iloc[probe_reflectivity_minimum_ind - self.points_below_SPR_min_ind:probe_reflectivity_minimum_ind + self.points_above_SPR_min_ind]
                                                                  })
                                               )

            # Next pair of probe point indices
            probepoint_index += 2

        # Create SPR vs TIR data frames
        injectionpoint_index = 0
        offset_SPR_sensorgram = self.sensorgram_data['SPR angle'] - self.sensorgram_data.loc[self.sensorgram_offset_ind, 'SPR angle']
        offset_TIR_sensorgram = self.sensorgram_data['TIR angle'] - self.sensorgram_data.loc[self.sensorgram_offset_ind, 'TIR angle']

        for reflectivity_index in range(int(len(self.injection_points) / 2)):

            self.SPR_vs_TIR_dfs.append(pd.DataFrame(data={'SPR angles': offset_SPR_sensorgram.iloc[self.injection_points[injectionpoint_index][0]:self.injection_points[injectionpoint_index + 1][0]],
                                                          'TIR angles': offset_TIR_sensorgram.iloc[self.injection_points[injectionpoint_index][0]:self.injection_points[injectionpoint_index + 1][0]],
                                                          })
                                       )

            # Next pair of injection point indices
            injectionpoint_index += 2

        return


def calculate_exclusion_height(exclusion_height_analysis_object_copy, buffer_or_probe_flag, data_frame_index):
    """
    This function calculates the exclusion height for a single injection step

    :param exclusion_height_analysis_object_copy: object containing all parameters and data
    :param buffer_or_probe_flag: either 'buffer' or 'probe'
    :param data_frame_index: index of dataframe and buffer or probe RI bulk value
    :return RI_results: list of calculated RI results for each height step
    """

    # Adapt new initial guess and bounds for refractive index range
    RI_results = []
    offset_results = []
    prism_k_results = []
    if not exclusion_height_analysis_object_copy.fit_offset:
        exclusion_new_fresnel_ini_guess = 1.38  # Swollen layer estimated hydration
        exclusion_new_fresnel_bounds = [1.0, 3.0]

    elif exclusion_height_analysis_object_copy.fit_offset and not exclusion_height_analysis_object_copy.fit_prism:
        exclusion_new_fresnel_ini_guess = exclusion_height_analysis_object_copy.fresnel_object.ini_guess[0:2]
        exclusion_new_fresnel_ini_guess[0] = 1.38  # Swollen layer estimated hydration
        exclusion_new_fresnel_bounds = [(1.0, -np.inf), (3.0, np.inf)]

    elif exclusion_height_analysis_object_copy.fit_offset and exclusion_height_analysis_object_copy.fit_prism:
        if len(exclusion_height_analysis_object_copy.fresnel_object.ini_guess) == 3:
            exclusion_new_fresnel_ini_guess = exclusion_height_analysis_object_copy.fresnel_object.ini_guess
        elif len(exclusion_height_analysis_object_copy.fresnel_object.ini_guess) == 2:
            exclusion_new_fresnel_ini_guess = [exclusion_height_analysis_object_copy.fresnel_object.ini_guess, 0.01]
        exclusion_new_fresnel_ini_guess[0] = 1.38  # Swollen layer estimated hydration
        exclusion_new_fresnel_bounds = [(1.0, -np.inf, 0), (3.0, np.inf, 0.1)]

    # Check if calculations should be performed on buffer or probe data
    if buffer_or_probe_flag == 'buffer':

        # Add bulk RI to layers
        refractive_indices = exclusion_height_analysis_object_copy.sensor_object.refractive_indices
        refractive_indices[-1] = exclusion_height_analysis_object_copy.buffer_bulk_RIs[data_frame_index]

        for height in exclusion_height_analysis_object_copy.height_steps:

            exclusion_height_analysis_object_copy.sensor_object.layer_thicknesses[-2] = height  # Surface layer height should be updated to current height step

            # Perform the fitting
            result = scipy.optimize.least_squares(fresnel_calculation,
                                                  exclusion_new_fresnel_ini_guess,
                                                  bounds=exclusion_new_fresnel_bounds,
                                                  kwargs={'fitted_layer_index': (-2, 2),  # Should always be the RI of the surface layer
                                                          'wavelength': exclusion_height_analysis_object_copy.sensor_object.wavelength,
                                                          'layer_thicknesses': exclusion_height_analysis_object_copy.sensor_object.layer_thicknesses,
                                                          'n_re': refractive_indices,
                                                          'n_im': exclusion_height_analysis_object_copy.sensor_object.extinction_coefficients,
                                                          'angles': exclusion_height_analysis_object_copy.buffer_reflectivity_dfs[data_frame_index]['angles'].to_numpy(),
                                                          'ydata': exclusion_height_analysis_object_copy.buffer_reflectivity_dfs[data_frame_index]['reflectivity'].to_numpy(),
                                                          'ydata_type': exclusion_height_analysis_object_copy.sensor_object.data_type,
                                                          'ydata_offset': exclusion_height_analysis_object_copy.fresnel_object.y_offset,
                                                          'polarization': exclusion_height_analysis_object_copy.polarization}
                                                  )
            # Collect the results from least_squares object
            RI_results.append(result['x'][0])

            if exclusion_height_analysis_object_copy.fit_offset and not exclusion_height_analysis_object_copy.fit_prism:
                offset_results.append(result['x'][1])

            elif exclusion_height_analysis_object_copy.fit_offset and exclusion_height_analysis_object_copy.fit_prism:
                offset_results.append(result['x'][1])
                prism_k_results.append(result['x'][2])

        return [RI_results, offset_results, prism_k_results]

    elif buffer_or_probe_flag == 'probe':

        # Add bulk RI to layers
        refractive_indices = exclusion_height_analysis_object_copy.sensor_object.refractive_indices
        refractive_indices[-1] = exclusion_height_analysis_object_copy.probe_bulk_RIs[data_frame_index]

        for height in exclusion_height_analysis_object_copy.height_steps:

            exclusion_height_analysis_object_copy.sensor_object.layer_thicknesses[-2] = height

            result = scipy.optimize.least_squares(fresnel_calculation,
                                                  exclusion_new_fresnel_ini_guess,
                                                  bounds=exclusion_new_fresnel_bounds,
                                                  kwargs={'fitted_layer_index': (-2, 2),  # Should always be the RI of the surface layer
                                                          'wavelength': exclusion_height_analysis_object_copy.sensor_object.wavelength,
                                                          'layer_thicknesses': exclusion_height_analysis_object_copy.sensor_object.layer_thicknesses,
                                                          'n_re': refractive_indices,
                                                          'n_im': exclusion_height_analysis_object_copy.sensor_object.extinction_coefficients,
                                                          'angles': exclusion_height_analysis_object_copy.probe_reflectivity_dfs[data_frame_index]['angles'].to_numpy(),
                                                          'ydata': exclusion_height_analysis_object_copy.probe_reflectivity_dfs[data_frame_index]['reflectivity'].to_numpy(),
                                                          'ydata_type': exclusion_height_analysis_object_copy.sensor_object.data_type,
                                                          'ydata_offset': exclusion_height_analysis_object_copy.fresnel_object.y_offset,
                                                          'polarization': exclusion_height_analysis_object_copy.polarization}
                                                  )
            # Collect the results from least_squares object
            RI_results.append(result['x'][0])

            if exclusion_height_analysis_object_copy.fit_offset and not exclusion_height_analysis_object_copy.fit_prism:
                offset_results.append(result['x'][1])

            elif exclusion_height_analysis_object_copy.fit_offset and exclusion_height_analysis_object_copy.fit_prism:
                offset_results.append(result['x'][1])
                prism_k_results.append(result['x'][2])

        return [RI_results, offset_results, prism_k_results]

    else:
        raise ValueError('Only buffer or probe allowed')


def exclusion_height_process(exclusion_height_analysis_object_copy, buffer_or_probe_flag, data_frame_index, connection):
    """
    This function initiates the calculations and sends back the result

    :param exclusion_height_analysis_object_copy: object containing all parameters and data
    :param buffer_or_probe_flag: either 'buffer' or 'probe'
    :param data_frame_index: index of dataframe
    :param connection: child pipe connection object from multiprocessing.Pipe()
    :return: None
    """

    result = calculate_exclusion_height(exclusion_height_analysis_object_copy, buffer_or_probe_flag, data_frame_index)
    connection.send(result)
    connection.close()


def process_all_exclusion_heights(exclusion_height_analysis_object, logical_cores):

    buffer_connections = []
    probe_connections = []
    buffer_processes = []
    probe_processes = []

    buffer_index = 0
    probe_index = 0
    process_step = 0
    required_processes = int(len(exclusion_height_analysis_object.injection_points) + len(exclusion_height_analysis_object.injection_points) / 2)

    while process_step < min(required_processes, logical_cores):

        # Setup buffer process
        buffer_parent_conn, buffer_child_conn = multiprocessing.Pipe()
        buffer_connections.append(buffer_parent_conn)
        buffer_process = multiprocessing.Process(target=exclusion_height_process, args=(copy.deepcopy(exclusion_height_analysis_object), 'buffer', buffer_index, buffer_child_conn))
        buffer_processes.append(buffer_process)
        buffer_process.start()
        buffer_index += 1
        process_step += 1

        # Setup probe process
        if buffer_index % 2 == 0:
            probe_parent_conn, probe_child_conn = multiprocessing.Pipe()
            probe_connections.append(probe_parent_conn)
            probe_process = multiprocessing.Process(target=exclusion_height_process, args=(copy.deepcopy(exclusion_height_analysis_object), 'probe', probe_index, probe_child_conn))
            probe_processes.append(probe_process)
            probe_process.start()
            probe_index += 1
            process_step += 1

    # If there are remaining processes, wait for the next two processes to finish and start new processes for the remaining steps
    process_index = 0
    while process_step < required_processes:
        buffer_processes[process_index].join()

        # Setup buffer process
        buffer_parent_conn, buffer_child_conn = multiprocessing.Pipe()
        buffer_connections.append(buffer_parent_conn)
        buffer_process = multiprocessing.Process(target=exclusion_height_process, args=(copy.deepcopy(exclusion_height_analysis_object), 'buffer', buffer_index, buffer_child_conn))
        buffer_processes.append(buffer_process)
        buffer_process.start()
        buffer_index += 1
        process_step += 1

        if buffer_index % 2 == 0:
            probe_processes[int(process_index/2)].join()

            # Setup probe process
            probe_parent_conn, probe_child_conn = multiprocessing.Pipe()
            probe_connections.append(probe_parent_conn)
            probe_process = multiprocessing.Process(target=exclusion_height_process, args=(copy.deepcopy(exclusion_height_analysis_object), 'probe', probe_index, probe_child_conn))
            probe_processes.append(probe_process)
            probe_process.start()
            probe_index += 1
            process_step += 1

        process_index += 1

    # Wait for each process to finish and collect and record results
    for process_index in range(len(buffer_processes)):

        # Check if user has aborted the analysis
        if exclusion_height_analysis_object.abort_flag:
            for buffer_process in buffer_processes:
                buffer_process.terminate()
            for probe_process in probe_processes:
                probe_process.terminate()

            exclusion_height_analysis_object.abort_flag = False

            return

        # Wait for processes to finish and collect results
        buffer_processes[process_index].join()
        full_buffer_result = buffer_connections[process_index].recv()
        buffer_RI_result = full_buffer_result[0]

        if process_index % 2 == 0:
            probe_processes[int(process_index/2)].join()
            full_probe_result = probe_connections[int(process_index/2)].recv()
            probe_RI_result = full_probe_result[0]

        if not exclusion_height_analysis_object.fit_offset:
            exclusion_height_analysis_object.d_n_pair_dfs[process_index] = pd.DataFrame(data={'height': exclusion_height_analysis_object.height_steps, 'buffer RI': buffer_RI_result, 'probe RI': probe_RI_result})
        elif exclusion_height_analysis_object.fit_offset and not exclusion_height_analysis_object.fit_prism:
            exclusion_height_analysis_object.d_n_pair_dfs[process_index] = pd.DataFrame(data={'height': exclusion_height_analysis_object.height_steps, 'buffer RI': buffer_RI_result, 'probe RI': probe_RI_result, 'buffer offsets': full_buffer_result[1], 'probe offsets': full_probe_result[1]})
        elif exclusion_height_analysis_object.fit_offset and exclusion_height_analysis_object.fit_prism:
            exclusion_height_analysis_object.d_n_pair_dfs[process_index] = pd.DataFrame(data={'height': exclusion_height_analysis_object.height_steps, 'buffer RI': buffer_RI_result, 'probe RI': probe_RI_result, 'buffer offsets': full_buffer_result[1], 'probe offsets': full_probe_result[1], 'buffer prism k': full_buffer_result[2], 'probe prism k': full_probe_result[2]})

        # Calculate exclusion height from buffer and probe height steps and RI result intersection and add to exclusion_height_analysis_object.all_exclusion_results
        for height_ind in range(len(exclusion_height_analysis_object.height_steps)):
            if buffer_RI_result[height_ind] < probe_RI_result[height_ind]:

                buffer_RI_zoom_range = np.linspace(buffer_RI_result[height_ind - 1], buffer_RI_result[height_ind], 200)
                probe_RI_zoom_range = np.linspace(probe_RI_result[height_ind - 1], probe_RI_result[height_ind], 200)
                height_zoom_range = np.linspace(exclusion_height_analysis_object.height_steps[height_ind - 1], exclusion_height_analysis_object.height_steps[height_ind], 200)

                # Zoom in on the intersection
                for index in range(200):
                    if buffer_RI_zoom_range[index] < probe_RI_zoom_range[index]:
                        exclusion_height_analysis_object.all_exclusion_results[0, process_index] = height_zoom_range[index]
                        exclusion_height_analysis_object.all_exclusion_results[1, process_index] = buffer_RI_zoom_range[index]
                        break

                # Stop looping through height steps
                break

    # Replace zeros with NaNs for non-intersecting results
    exclusion_height_analysis_object.all_exclusion_results = np.where(exclusion_height_analysis_object.all_exclusion_results == 0, np.nan, exclusion_height_analysis_object.all_exclusion_results)

    return


def add_sensor_backend(session_object, data_path_, default_sensor_values, sensor_metal='Au'):

    """
    Adds sensor objects to a session object.
    :return: a sensor object
    """
    session_object.sensor_ID_count += 1
    sensor_object = Sensor(data_path_, session_object.sensor_ID_count, default_sensor_values, sensor_metal=sensor_metal)
    session_object.sensor_instances[session_object.sensor_ID_count] = sensor_object

    return sensor_object


def copy_sensor_backend(session_object, sensor_object):

    """
    Copies sensor object to a session object.
    :return: a sensor object
    """
    session_object.sensor_ID_count += 1
    copied_sensor_object = copy.deepcopy(sensor_object)
    copied_sensor_object.object_id = session_object.sensor_ID_count
    session_object.sensor_instances[session_object.sensor_ID_count] = copied_sensor_object

    return copied_sensor_object


def add_fresnel_model_object(session_object, sensor_object, data_path_, reflectivity_df_, object_name_):
    """
    Adds analysis objects to a session object.
    :return: an analysis object
    """
    session_object.fresnel_analysis_ID_count += 1
    analysis_object = FresnelModel(session_object, sensor_object, data_path_, reflectivity_df_, session_object.fresnel_analysis_ID_count, object_name_)
    session_object.fresnel_analysis_instances[session_object.fresnel_analysis_ID_count] = analysis_object

    return analysis_object


def add_exclusion_height_object(session_object, fresnel_object, sensorgram_df_, data_path_, object_name_):
    """
    Adds analysis objects to a session object.
    :return: an analysis object
    """
    session_object.exclusion_height_analysis_ID_count += 1
    analysis_object = ExclusionHeight(session_object, fresnel_object, sensorgram_df_, data_path_, session_object.exclusion_height_analysis_ID_count, object_name_)
    session_object.exclusion_height_analysis_instances[session_object.exclusion_height_analysis_ID_count] = analysis_object

    return analysis_object



# This file contains utility functions

import numpy as np
import tkinter
from tkinter.filedialog import askopenfilename, askopenfilenames, askdirectory, asksaveasfilename
import pandas as pd
import re
from fresnel_transfer_matrix import TIR_determination


def select_folder(prompt, prompt_folder=None):
    root = tkinter.Tk()
    root.attributes("-topmost", 1)
    root.withdraw()
    selected_folder = askdirectory(title=prompt, parent=root, initialdir=prompt_folder)
    root.destroy()
    return selected_folder


def select_file(prompt, prompt_folder=None, file_types=[('Pickle files', '*.pickle')]):
    root = tkinter.Tk()
    root.attributes("-topmost", 1)
    root.withdraw()
    selected_file = askopenfilename(title=prompt, filetypes=file_types, initialdir=prompt_folder, parent=root)
    root.destroy()
    return selected_file


def select_files(prompt, prompt_folder=None, file_types=[('Pickle files', '*.pickle')]):
    root = tkinter.Tk()
    root.attributes("-topmost", 1)
    root.withdraw()
    selected_files = askopenfilenames(title=prompt, filetypes=file_types, initialdir=prompt_folder, parent=root)
    root.destroy()
    return selected_files


def save_file(prompt, prompt_folder=None, file_types=[('CSV files', '*.csv')], default_extension='.csv'):
    root = tkinter.Tk()
    root.attributes("-topmost", 1)
    root.withdraw()
    save_file = asksaveasfilename(title=prompt, filetypes=file_types, defaultextension=default_extension, initialdir=prompt_folder, parent=root)
    root.destroy()
    return save_file


def load_csv_data(path=False, default_data_folder=None, prompt='Select the measurement data file (.csv)'):
    if not path:
        print(prompt)
        data_path_ = select_file('Select the measurement data file (.csv)', prompt_folder=default_data_folder, file_types=[('CSV files', '*.csv')])
    else:
        data_path_ = path

    #  Determine the scanning speed/step length if present in the file
    try:
        with open(data_path_, 'r') as file:
            step_length_pattern = re.compile(r'=\d{1,2}')
            scanspeed = int(step_length_pattern.search(file.readline()).group().strip('='))

    except AttributeError:  # I think .group().strip() should return AttributeError if .search() returns None
        scanspeed = 5  # Assuming medium scanspeed if legacy spr2 to csv converter was used


    # Load in the measurement data from a .csv file
    data_frame_ = pd.read_csv(data_path_, delimiter=';', skiprows=1, header=None)
    time_df = data_frame_.iloc[1:, 0]
    angles_df = data_frame_.iloc[0, 1:]
    ydata_df = data_frame_.iloc[1:, 1:]

    # Select last scan as default reflectivity plot
    reflectivity_df_ = pd.DataFrame(data={'angles': angles_df, 'ydata': ydata_df.iloc[-1, :]})

    return data_path_, scanspeed, time_df, angles_df, ydata_df, reflectivity_df_


def calculate_sensorgram(time, angles, ydata, SPR_TIR_fitting_parameters):

    # Convert dataframes to numpy ndarrays
    time = time.to_numpy()
    angles = angles.to_numpy()
    ydata = ydata.to_numpy()

    # Calculating SPR and TIR angles
    sensorgram_SPR_angles = np.empty(len(ydata))*np.nan
    sensorgram_SPR_fit_y = [pd.Series(np.empty(SPR_TIR_fitting_parameters['SPR fit points'])*np.nan)]*len(ydata)
    sensorgram_SPR_fit_x = [pd.Series(np.empty(SPR_TIR_fitting_parameters['SPR fit points'])*np.nan)]*len(ydata)

    sensorgram_TIR_angles = np.empty(len(ydata))*np.nan
    TIR_deriv_points = len(angles[(angles >= SPR_TIR_fitting_parameters['TIR range'][0]) & (angles <= SPR_TIR_fitting_parameters['TIR range'][1])])
    sensorgram_TIR_deriv_x = [pd.Series(np.empty(TIR_deriv_points)*np.nan)]*len(ydata)
    sensorgram_TIR_deriv_y = [pd.Series(np.empty(TIR_deriv_points)*np.nan)]*len(ydata)
    sensorgram_TIR_deriv_fit_x = [pd.Series(np.empty(SPR_TIR_fitting_parameters['TIR fit points']) * np.nan)] * len(ydata)
    sensorgram_TIR_deriv_fit_y = [pd.Series(np.empty(SPR_TIR_fitting_parameters['TIR fit points']) * np.nan)] * len(ydata)

    for ind, val in enumerate(time):
        reflectivity_spectrum = ydata[ind-1, :]
        min_index = np.argmin(reflectivity_spectrum)

        # SPR angles
        try:
            y_selection = reflectivity_spectrum[min_index - SPR_TIR_fitting_parameters['sensorgram_angle_range_points'][0]:min_index + SPR_TIR_fitting_parameters['sensorgram_angle_range_points'][1]]

            polynomial = np.polyfit(angles[min_index - SPR_TIR_fitting_parameters['sensorgram_angle_range_points'][0]:min_index + SPR_TIR_fitting_parameters['sensorgram_angle_range_points'][1]],
                                    y_selection, 3)
            x_selection = np.linspace(angles[min_index - SPR_TIR_fitting_parameters['sensorgram_angle_range_points'][0]],
                                      angles[min_index + SPR_TIR_fitting_parameters['sensorgram_angle_range_points'][1]], SPR_TIR_fitting_parameters['SPR fit points'])
            y_polyfit = np.polyval(polynomial, x_selection)
            y_fit_min_ind = np.argmin(y_polyfit)

            sensorgram_SPR_angles[ind-1] = x_selection[y_fit_min_ind]
            sensorgram_SPR_fit_x[ind-1] = pd.Series(x_selection)
            sensorgram_SPR_fit_y[ind-1] = pd.Series(y_polyfit)

        except:
            print('No SPR minimum found. Skipping measurement time point {}...'.format(val))


        # TIR angles
        try:
            TIR_theta, TIR_xdata_filtered, deriv_ydata, TIR_theta_fit_x, TIR_theta_fit_y  = TIR_determination(angles, reflectivity_spectrum, SPR_TIR_fitting_parameters)
            sensorgram_TIR_angles[ind-1] = TIR_theta
            sensorgram_TIR_deriv_x[ind-1] = pd.Series(TIR_xdata_filtered)
            sensorgram_TIR_deriv_y[ind-1] = pd.Series(deriv_ydata)
            sensorgram_TIR_deriv_fit_x[ind-1] = pd.Series(TIR_theta_fit_x)
            sensorgram_TIR_deriv_fit_y[ind-1] = pd.Series(TIR_theta_fit_y)

        except:
            print('No TIR found. Skipping measurement time point {}...'.format(val))


    sensorgram_df = pd.DataFrame(data={'time': time, 'SPR angle': sensorgram_SPR_angles, 'TIR angle': sensorgram_TIR_angles, 'SPR fit x': sensorgram_SPR_fit_x, 'SPR fit y': sensorgram_SPR_fit_y, 'TIR deriv x': sensorgram_TIR_deriv_x, 'TIR deriv y': sensorgram_TIR_deriv_y, 'TIR deriv fit x': sensorgram_TIR_deriv_fit_x, 'TIR deriv fit y': sensorgram_TIR_deriv_fit_y})

    return sensorgram_df

